import os
import PIL

import torch
from torch.utils.data import Dataset


class ImageDatasetFromDF(Dataset):
    r"""Creates image dataset from a `pandas.DataFrame` instance.

    Parameters
    ----------
    - `root_directory`: str
        directory for image files
    - `dataframe`: pandas.DataFrame
        dataframe object with image filename and labels.
    - `image_file_col`: str
        column name of image filename.
    - `label_col`: str
        column name of label column
    - `mapping`: dict
        labels to class idx mapping
    """

    def __init__(
        self,
        root_directory,
        dataframe,
        image_file_col,
        label_col,
        mapping,
        transforms=None,
    ):
        self.df = dataframe.reset_index(drop=True)
        self.dir = root_directory
        self.tfrm = transforms
        self.labels = label_col
        self.ifl = image_file_col

        self.class_to_label = mapping

    def __len__(self):
        return len(self.df)

    def __getitem__(self, index):
        filename = os.path.join(str(self.dir), str(self.df.loc[index, self.ifl]))
        labels = torch.tensor(
            self.class_to_label[self.df.loc[index, self.labels]], dtype=torch.float32
        )
        image = PIL.Image.open(filename)
        if self.tfrm:
            image = self.tfrm(image)

        return image, labels


class ImageDatabunch(Dataset):
    r"""Creates an image dataset from `img_dir` parameter. Useful for making prediction on a image folder.
    Parameters
    ----------
    - `root_directory`: str
        directory for image files
    """

    def __init__(self, img_dir, transforms=None):
        self.dir = img_dir
        self.image_files = [
            f
            for f in os.listdir(img_dir)
            if f.endswith(".png") or f.endswith(".jpg") or f.endswith(".jpeg")
        ]
        self.tfrm = transforms

    def __len__(self):
        return len(self.image_files)

    def __getitem__(self, index):
        filename = os.path.join(str(self.dir), str(self.image_files[index]))
        image = PIL.Image.open(filename)
        if self.tfrm:
            image = self.tfrm(image)

        return image, str(self.image_files[index])



import os
import math
import logging
from tqdm import *
import urllib.request
from typing import Union, List
from selenium import webdriver
from webdriver_manager.chrome import ChromeDriverManager


def init_driver(fetcher):
    options = webdriver.ChromeOptions()
    options.add_argument("headless")
    tab = webdriver.Chrome(
        ChromeDriverManager(log_level=logging.WARNING).install(), options=options
    )

    def fetcher_with_tab(*args, **kwargs):
        return fetcher(*args, **kwargs, driver=tab)

    return fetcher_with_tab


@init_driver
def google_images_scraper(
    search_term: str,
    min_image_count: int = 10,
    driver: webdriver.chrome.webdriver.WebDriver = None,
) -> list:
    base_url = "https://www.google.com/imghp?hl=en"
    driver.get(base_url)
    search_form = driver.find_element_by_name("q")
    search_form.send_keys(search_term)
    search_form.submit()
    urls = []
    last_height = driver.execute_script("return document.body.scrollHeight")

    reached_page_end = False
    while len(urls) < min_image_count:
        driver.execute_script("window.scrollTo(0, document.body.scrollHeight);")
        urls = driver.execute_script(
            "return Array.from(document.querySelectorAll('.rg_i')).map(el=> el.hasAttribute('data-src')?el.getAttribute('data-src'):el.getAttribute('data-iurl'));"
        )

        while None in urls:
            urls.remove(None)

        new_height = driver.execute_script("return document.body.scrollHeight")
        if last_height == new_height:
            break
        else:
            last_height = new_height

    return urls


def download_images_from_url(
    urls: Union[str, List[str]],
    image_tag: str,
    format: str = "png",
    save_location: str = ".",
) -> None:
    dir = os.path.join(save_location, image_tag)
    if not os.path.exists(dir):
        os.mkdir(dir)

    if isinstance(urls, str):
        urllib.request.urlretrieve(urls, os.path.join(dir, "0001" + "." + format))
        print(f"Downloaded 1 image in {dir}.")

    elif isinstance(urls, list):
        pad = math.ceil(math.log10(len(urls)))
        for idx in tqdm(range(len(urls))):
            urllib.request.urlretrieve(
                urls[idx], os.path.join(dir, str(idx + 1).zfill(pad) + "." + format)
            )
        print(f"Downloaded {len(urls)} images in {dir}.")

    else:
        raise Exception('Invalid argument type for "urls"')



import torch
import torch.nn as nn
import torch.nn.functional as F


class GenericModel(nn.Module):
    r"""A generic model class which accepts different criterion functions as parameter::

        import torch.nn as nn
        import torch.nn.functional as F

        import torchlit.model as model

        class Net(model.Model):
            def __init__(self):
                super(Net, self).__init__(F.cross_entropy, record=True, verbose=True)
                self.conv1 = nn.Conv2d(1, 20, 5)
                self.conv2 = nn.Conv2d(20, 20, 5)

            def forward(self, x):
                x = F.relu(self.conv1(x))
                return F.relu(self.conv2(x))


    `record`:: `bool`:: `Optional`: Record the history over course of training.
    `verbose`:: `bool`:: `Optional`: Display information over epochs.
    """

    results = {
        "train": {"loss": []},
        "val": {"loss": [], "acc": []},
    }

    epoch = 0
    history = []

    def __init__(self, criterion, record: bool = True, verbose: bool = True):
        super(GenericModel, self).__init__()

        self.criterion = criterion
        self.record = record
        self.verbose = verbose

    def train_step(self, xb):
        data, labels = xb
        out = self(data)
        loss = self.criterion(out, labels)

        self.results["train"]["loss"].append(loss.cpu().item())

        return loss

    def val_step(self, xb):
        data, labels = xb
        out = self(data)
        loss = self.criterion(out, labels)
        acc = self.accuracy(out, labels)

        self.results["val"]["loss"].append(loss.cpu().item())
        self.results["val"]["acc"].append(acc.cpu().item())

    def epoch_end(self):
        train_loss = torch.tensor(self.results["train"]["loss"]).mean().item()
        val_loss = torch.tensor(self.results["val"]["loss"]).mean().item()
        val_acc = torch.tensor(self.results["val"]["acc"]).mean().item()

        if self.record:
            self.history.append(
                {
                    "epoch": self.epoch,
                    "train_loss": train_loss,
                    "val_loss": val_loss,
                    "val_acc": val_acc,
                }
            )
        if self.verbose:
            print(
                f"Epoch [{self.epoch}]: train_loss: {train_loss}, val_loss: {val_loss}, val_acc: {val_acc}"
            )

        self.reset_results()
        self.epoch += 1

    def reset_results(self):
        self.results = {
            "train": {"loss": []},
            "val": {"loss": [], "acc": []},
        }

    @staticmethod
    def accuracy(outputs, labels):
        _, preds = torch.max(outputs, dim=1)
        return torch.tensor(torch.sum(preds == labels).item() / len(preds))


class ImageClassificationModel(GenericModel):
    """
    Image Classification Model for Mulitclass Classification
    """

    def __init__(self, record=True, verbose=True):
        super().__init__(F.cross_entropy, record, verbose)



from setuptools import find_packages, setup

with open("README.md") as f:
    long_description = f.read()


if __name__ == "__main__":
    setup(
        name="torchlit",
        version="0.1.1",
        description="torchlit - thin wrappers for Pytorch",
        long_description=long_description,
        long_description_content_type="text/markdown",
        author="Himanshu Dutta",
        author_email="meet.himanshu.dutta@gmail.com",
        url="https://github.com/himanshu-dutta/torchlit",
        license="MIT License",
        packages=find_packages(),
        include_package_data=True,
        install_requires=[
            "google-cloud-storage>=1.35.1",
            "selenium>=3.141.0",
            "torch>=1.6.0",
            "torchaudio>=0.7.2",
            "torchvision>=0.8.2",
            "tqdm>=4.55.0",
            "webdriver-manager>=3.2.2",
        ],
        platforms=["linux", "unix"],
        python_requires=">3.5.2",
    )